#Offline diffraction simulation scripts from pisha2020approximate
#Copyright (C) 2020 Louis Pisha <lpisha@eng.ucsd.edu> / Sonic Arts Research & Development <sonicarts.ucsd.edu> / UC San Diego / The Regents of the University of California 
#See LICENSE.txt for licensing information

#Biot-Tolstoy-Medwin Edge Diffraction Model
#Based on svensson1999analytic and svensson2006edge
#Equation numbers are from svensson1999analytic by default

from common import *
from math import *

# Expensive numerical integration
from scipy.integrate import quad

def cheapnumint(f, lo, hi):
    assert(lo < hi)
    #fstart = f(lo)
    fmid = f((lo+hi)/2.0)
    #fend = f(hi)
    #assert(abs(fstart - fend) < 1)
    return fmid * (hi-lo), None

# Integrand of equation 33
def btm_z(thetai, ri, zi, thetad, rd, zd, v, z):
    dzi = z - zi
    dzd = z - zd
    liz = sqrt(dzi * dzi + ri * ri)
    ldz = sqrt(dzd * dzd + rd * rd)
    #
    # Equation 16
    # sinalphaz = dzi / liz
    # singammaz = dzd / ldz
    # cosalphaz = ri / liz
    # cosgammaz = rd / ldz
    # etaz = acosh((1.0 + sinalphaz * singammaz) / (cosalphaz * cosgammaz))
    # coshvetaz_2 = cosh(v * etaz)
    #
    # Equation 36 and following
    y = (liz * ldz + dzi * dzd) / (ri * rd)
    A = y + sqrt(y * y - 1)
    Apow = pow(A, v)
    coshvetaz = 0.5 * (Apow + (1.0 / Apow))
    # assert(abs(coshvetaz - coshvetaz_2) < 1e-6)
    #
    # Equations 2-3
    def B(ang):
        return sin(v * ang) / (coshvetaz - cos(v * ang))
    Btotal = (B(pi + (thetad - thetai)) +
              B(pi - (thetad - thetai)) +
              B(pi + (thetad + thetai)) +
              B(pi - (thetad + thetai)) )
    return -v * Btotal / (4.0 * pi * liz * ldz)

# Computation of limits of equation 33 and integration
def btm_dn(thetai, ri, zi, li, thetad, rd, zd, ld, v, zedgelo, zedgehi, splspermeter, goodnumint, n):
    def zn(zi, zd, li, ld, delta):
        kq = li * li - ld * ld - delta * delta
        aq = (zi - zd) ** 2 - delta * delta
        bq = 2.0 * delta * delta * zd - kq * (zi - zd)
        cq = 0.25 * kq * kq - delta * delta * ld * ld
        bq /= aq
        cq /= aq
        sq = bq * bq - 4.0 * cq
        if sq < 0.0:
            return (None, None)
        sq = sqrt(sq)
        return ((-bq + sq) * 0.5, (-bq - sq) * 0.5)
    (zn1p, zn1m) = zn(zi, zd, li, ld, (n - 0.5) / splspermeter)
    (zn2p, zn2m) = zn(zi, zd, li, ld, (n + 0.5) / splspermeter)
    #print("zn2m " + str(zn2m) + " zn1m " + str(zn1m) + " zn1p " + str(zn1p) + " zn2p " + str(zn2p))
    if not zn2p:
        # Both limits of integration are imaginary
        # Entire sample has no existing edge
        #print('No edge for sample n = ' + str(n))
        return 0.0
    if not zn1p:
        # Only lower limit of integration is imaginary
        # Instead start integrating at apex point
        #print('At apex point for sample n = ' + str(n))
        zn1p = 0.0
        zn1m = 0.0
    # Check the ranges against the boundaries of the edge.
    # The two ranges are (from negative z to positive z):
    # [zn2m, zn1m]
    # [zn1p, zn2p]
    assert(zn2m <= zn1m)
    assert(zn1m <= zn1p)
    assert(zn1p <= zn2p)
    if zedgelo > zn2m: zn2m = zedgelo
    if zedgehi < zn1m: zn1m = zedgehi
    if zedgelo > zn1p: zn1p = zedgelo
    if zedgehi < zn2p: zn2p = zedgehi
    # Integrate two ranges, if each one is valid
    def btm_integrand(z):
        return btm_z(thetai, ri, zi, thetad, rd, zd, v, z)
    numint = quad if goodnumint else cheapnumint
    ret = 0.0
    onevalid = False
    if zn2m < zn1m:
        ret += numint(btm_integrand, zn2m, zn1m)[0]
        onevalid = True
    if zn1p < zn2p:
        ret += numint(btm_integrand, zn1p, zn2p)[0]
        onevalid = True
    if not onevalid:
        #print('Integration off ends of edge on both sides')
        #This is common for finite edge lengths
        pass
    return ret
    
    
def btm_ir(thetai, ri, zi, li, thetad, rd, zd, ld, v, zedgelo, zedgehi, splspermeter, goodnumint, nir):
    n0 = int(floor(splspermeter * (li + ld)))
    ir = np.zeros((nir,))
    for i in range(nir):
        n = n0 + i
        ir[i] = btm_dn(thetai, ri, zi, li, thetad, rd, zd, ld, v, zedgelo, zedgehi, splspermeter, goodnumint, n)
    return ir, n0

import numpy as np

def pl_firstnonzero(ir, n0, splspermeter):
    # Compute path length based on the first nonzero sample of the impulse response
    w = np.argwhere(abs(ir) > 1e-7)
    if len(w) == 0:
        return None
    s = n0 + w[0].item()
    #print('firstnonzero: sample ' + str(s) + ', result ' + str(s / splspermeter))
    return s / splspermeter

def pl_meanenergy(ir, n0, splspermeter):
    # Compute path length based on the time when half the energy of the IR has passed
    cum_energy = np.cumsum(np.square(ir))
    total_energy = cum_energy[-1].item()
    if total_energy == 0.0:
        return None
    w = np.argwhere(cum_energy >= 0.5 * total_energy)
    s = n0 + w[0].item()
    #print('meanenergy: sample ' + str(s) + ', result ' + str(s / splspermeter))
    return s / splspermeter


def call_btm(verts, tris, edges, ps, pk, always, btmparams):
    splspermeter = btmparams['fs'] / btmparams['c']
    res = []
    pls_nz = []
    pls_me = []
    for edge in edges:
        ri, rd, zi, zd, li, ld, phii, apex, inedge, zedgelo, zedgehi, thetai, thetad, v \
            = edge_params(verts, tris, edge, ps, pk)
        if thetad - thetai < pi and not always:
            # Not in shadow region of this edge
            print('BTM not doing non-shadowed diffraction')
            continue
        ir, n0 = btm_ir(thetai, ri, zi, li, thetad, rd, zd, ld, v, 
            zedgelo, zedgehi, splspermeter, btmparams['goodnumint'], btmparams['irlen'])
        res.append((ir, n0))
        pls_nz.append(pl_firstnonzero(ir, n0, splspermeter))
        pls_me.append(pl_meanenergy(ir, n0, splspermeter))
    if not SegIntersectAnyTriangle(verts, tris, ps, pk):
        #print('BTM direct path open')
        directlen = norm(np.array(pk) - np.array(ps))
        res.append(([1.0/directlen], floor(splspermeter * directlen))) #Impulse response is one nonzero
        #print('Direct len: sample ' + str(floor(splspermeter * directlen)) + ', result ' + str(directlen))
        pls_nz.append(directlen)
        pls_me.append(directlen)
    totalirlen = btmparams['totalirlen']
    totalir = np.zeros((totalirlen,))
    if len(res) == 0:
        print('Warning, no impulse responses for BTM')
    for result in res:
        ir = result[0]
        n0 = result[1]
        if n0 < 0:
            raise RuntimeError('n0 must be positive')
        if n0 + len(ir) > totalirlen:
            print('Warning, BTM IR from ' + str(n0) + ' to ' + str(n0 + len(ir)) 
                + ' getting cut off at ' + str(totalirlen))
        if n0 < totalirlen:
            count = min(len(ir), totalirlen-n0)
            #print(str(totalir.shape) + ' ' + str(ir.shape))
            totalir[n0:n0+count] = totalir[n0:n0+count] + ir[0:count]
    totalfd = np.fft.rfft(totalir)
    totalfd = abs(totalfd)
    freqs = np.fft.rfftfreq(totalirlen, 1.0/btmparams['fs'])
    pls_nz.append(pl_firstnonzero(totalir, 0, splspermeter))
    pls_me.append(pl_meanenergy(totalir, 0, splspermeter))
    pls_nz = [p for p in pls_nz if p is not None]
    pls_me = [p for p in pls_me if p is not None]
    assert(len(pls_nz) > 0)
    assert(len(pls_me) > 0)
    return totalfd, freqs, pls_nz, pls_me
