#Offline diffraction simulation scripts from pisha2020approximate
#Copyright (C) 2020 Louis Pisha <lpisha@eng.ucsd.edu> / Sonic Arts Research & Development <sonicarts.ucsd.edu> / UC San Diego / The Regents of the University of California 
#See LICENSE.txt for licensing information

# VDaT approximating BTM - Random disks functions.

import random, pickle
import math
import numpy as np
import os

from common import *
from btm import *

def rand_obstacle(seed, rin, rout, outside):
    random.seed(seed)
    okay = False
    while not okay:
        okay = True
        verts = [] if outside else [[0, 0, 0]]
        tris = []
        edges = []
        nseg = random.randrange(3, random.randrange(5, 12))
        angle = 0.0
        s = 0
        while s < nseg and okay:
            # Edge/triangle
            s2 = s+1 if s < nseg-1 else 0
            if outside:
                tris.append([s*2, s*2+1, s2*2])
                tris.append([s*2+1, s2*2, s2*2+1])
                edges.append({'i1': s*2, 'i2': s2*2, 'tis': [s*2], 'angle': math.tau})
            else:
                tris.append([0, s+1, s2+1])
                edges.append({'i1': s+1, 'i2': s2+1, 'tis': [s], 'angle': math.tau})
            # Vertex
            if angle > math.tau - 0.1:
                okay = False
                break
            r = random.uniform(rin, rout)
            verts.append([r*cos(angle), r*sin(angle), 0])
            if outside:
                verts.append([100*cos(angle), 100*sin(angle), 0])
            da = math.tau / nseg
            angle += random.uniform(da * 0.5, da * 1.5)
            s += 1
        if not okay:
            continue
        # Validate that edges are outside inner radius
        e = 0
        while e < len(edges) and okay:
            if norm(upright(verts[edges[e]['i1']], verts[edges[e]['i2']], [0,0,0], False)) <= rin:
                okay = False
                break
            e += 1
    return verts, tris, edges


class BTMMonteCarlo():
    def __init__(self, zi, zd, rin, rout, runname, color='blue', seed=None, outside=False,
        c=345.0, fs=160000, irlen=512, totalirlen=16384, goodnumint=False):
        self.zi = zi
        self.zd = -zi if zd is None else zd
        self.rin = rout * 0.5 if rin is None else rin
        self.rout = rout
        self.runname = runname
        self.color = color
        self.seed = seed if seed else random.randrange(0, 1000000000)
        self.outside = outside
        self.params = {'fs': fs, 'c': c, 'irlen': irlen, 'totalirlen': totalirlen, 'goodnumint': goodnumint}
        self.another(reset=True)
    
    def another(self, count=1, reset=False):
        for c in range(count):
            self.lastverts, self.lasttris, self.lastedges = rand_obstacle(
                self.seed, self.rin, self.rout, self.outside)
            self.lastres, self.freqs, _, _ = call_btm(
                self.lastverts, self.lasttris, self.lastedges, 
                [0,0,self.zi], [0,0,self.zd], True, self.params)
            self.seed += 1
            if c == 0 and reset:
                self.summags = np.array(self.lastres)
                self.nummags = 1.0
            else:
                self.summags += np.array(self.lastres)
                self.nummags += 1.0
    
    def get_freqs(self):
        return self.freqs
    
    def get_lastobject(self):
        return self.lastverts, self.lasttris, self.lastedges
    
    def get_lastres(self):
        return todb(self.lastres)
        
    def get_avg(self):
        return todb(self.summags / self.nummags)
        
    def get_rs(self):
        return self.rin, self.rout
        
    def save(self):
        os.makedirs('out', exist_ok=True)
        label = self.runname + '_{:.3f}'.format(self.rout)
        filename = 'out/' + label.replace('.', 'p') + '.pickle'
        r = {'freqs': self.freqs, 'mags': self.get_avg(), 
            'zi': self.zi, 'zd': self.zd, 'rin': self.rin, 'rout': self.rout,
            'outside': self.outside, 'color': self.color, 'label': label}
        with open(filename, 'wb') as p:
            pickle.dump(r, p)
            print('Saved to ' + filename)
