#Offline diffraction simulation scripts from pisha2020approximate
#Copyright (C) 2020 Louis Pisha <lpisha@eng.ucsd.edu> / Sonic Arts Research & Development <sonicarts.ucsd.edu> / UC San Diego / The Regents of the University of California 
#See LICENSE.txt for licensing information

from math import *
import cmath
import numpy as np
from numpy.linalg import norm

################################################################################
# Math
################################################################################

def pm(plus, x):
    return x if plus else -x
    
def degrees(x):
    return x * 180.0 / pi
    
def radians(x):
    return x * pi / 180.0

def normalize(v):
    n = norm(v)
    assert(n > 1e-7)
    return v / n

def thirdvertex(t, i1, i2):
    return [i for i in t if i not in [i1, i2]][0]

def upright(s, e, p, get_d = False):
    s, e, p = np.array(s), np.array(e), np.array(p)
    ves, vps = e - s, p - s
    ves = normalize(ves)
    d = np.dot(ves, vps)
    proj = ves * d
    ret = vps - proj
    return (ret, d) if get_d else ret

def triuprightn(verts, t, i1, i2):
    return normalize(upright(verts[i1], verts[i2], verts[thirdvertex(t, i1, i2)]))

def acos3(d_zero, d_90, d):
    ret = acos(min(1.0, max(-1.0, np.dot(d_zero, d))))
    if np.dot(d_90, d) < 0:
        ret = tau - ret
    return ret

def todb(res):
    try:
        _ = res[0]
        return 20.0 * np.log10(np.array(res) + 1e-6)
    except (IndexError, TypeError) as e:
        try:
            return 20.0 * log10(res)
        except TypeError as _no:
            raise e

def audible_only(freqs, target = None):
    where_freqs_in_range = np.logical_and(20.0 < freqs, freqs <= 20000.0)
    ret = np.extract(where_freqs_in_range, target if target is not None else freqs)
    assert(len(ret) >= 1)
    return ret

def FW_LSD(f, a, b, alreadydb):
    assert(len(f) == len(a) == len(b))
    if not alreadydb:
        a = todb(a)
        b = todb(b)
    f = audible_only(f)
    a = audible_only(f, a)
    b = audible_only(f, b)
    assert(len(a) == len(b))
    return sqrt(np.sum(np.square(a-b) / f) / np.sum(1.0 / f))

################################################################################
# GUI
################################################################################

# From https://stackoverflow.com/questions/37765197/darken-or-lighten-a-color-in-matplotlib
def adjust_lightness(color, amount=0.8):
    import matplotlib.colors as mc
    import colorsys
    try:
        c = mc.cnames[color]
    except:
        c = color
    c = colorsys.rgb_to_hls(*mc.to_rgb(c))
    return colorsys.hls_to_rgb(c[0], max(0, min(1, amount * c[1])), max(0.0, min(1.0, amount * c[2])))
    
def rand_color(other_colors=None):
    maxdist = 0.5
    while True:
        ret = np.random.random((3,))
        maxdist *= 0.9
        if not(0.4 < norm(ret) < 1.5):
            continue
        if other_colors is not None and any([norm(c - ret) < maxdist for c in other_colors]):
            continue
        return ret

def annotate_number(n):
    return '-inf' if n < -10000.0 else 'inf' if n > 10000.0 else str(round(n, 3))

def fix_widget(widget, isCheckButtons):
    ax = widget.ax
    invxfrm = ax.transAxes.inverted()
    unitx = invxfrm.transform((1, 0))[0] - invxfrm.transform((0, 0))[0]
    unity = invxfrm.transform((0, 1))[1] - invxfrm.transform((0, 0))[1]
    if isCheckButtons:
        bs = widget.lines
        cs = widget.rectangles
    else:
        bs = [None] * len(widget.labels)
        cs = widget.circles
    for lbl, b, c in zip(widget.labels, bs, cs):
        lbl.set_x(unitx * 22)
        dy_orig = 1.0 / (len(widget.labels) + 1)
        y_center = c.xy[1] + dy_orig / 3.0 if isCheckButtons else c._center[1] + dy_orig / 20.0
        x = unitx * 6
        y = y_center - unity * 6
        w = unitx * 12
        h = unity * 12
        if isCheckButtons:
            b[0].set_data([x, x+w], [y, y+h])
            b[1].set_data([x+w, x], [y, y+h])
            c.set_bounds(x, y, w, h)
        else:
            c._center, c.width, c.height, c.stale = (x+w/2, y+h/2), w, h, True


################################################################################
# Parameters
################################################################################

def defaultvdatparams():
    return {'subpath0': True,
            'subpath1': False,
            'subpath2': False,
            'subpath3': False,
            'dbdechf': 5.0,
            'dbdeclf': 1.0}
            
def defaultbtmparams():
    return {'fs': 44100,
            'irlen': 4096,
            'totalirlen': 16384,
            'goodnumint': False} 

def examplegeom():
    return {'sx': -1.0,
            'sy': -1.0,
            'kx': 1.0,
            'ky': -1.0,
            'edges': [
                {'y': 0.0, 'extent': -100000, 'widthlo': -100000, 'widthhi': 100000}
            ]}

def defaultringdefs(nrings, nangles, fs, c):
    ringdefs = []
    def add_rs(rs, nangles=nangles):
        for r in rs:
            rad = c / f
            ringdefs.append({'f': f, 'r': rad, 'k': 2.0 * pi / rad,
                'nangles': nangles, 'showsps': True})
    assert(nrings >= 1)
    f = (20480.0 / float(1 << nrings)) if nrings >= 9 else (640.0 / float(1 << (nrings >> 1)))
    for r in range(nrings):
        add_rs([f])
        f *= 2.0
    add_rs([fs / 2.0], 1)
    return ringdefs

def setup_ax(ax):
    ax.clear()
    ax.set_xscale('log')
    ax.set_xlim(20, 20000)
    ax.set_ylim(-60, 20)
    #ax.plot([20, 20], [-100, 100], linestyle='--', color='lightgrey')
    #ax.plot([20000, 20000], [-100, 100], linestyle='--', color='lightgrey')
    ax.set_xlabel('Frequency (Hz)')
    ax.set_ylabel('Amplitude response (dB)')

################################################################################
# Core
################################################################################

# Bias: make the triangle slightly larger or smaller (by 1 micron), to make edge
# cases more predictable.
def IsectTriSeg(tvs, s, e, bias='larger'):
    a = np.array(tvs[0])
    side1 = np.array(tvs[1]) - a
    side2 = np.array(tvs[2]) - a
    s, e = np.array(s), np.array(e)
    d = e - s
    trinorm = np.cross(side1, side2)
    det_M = -np.dot(d, trinorm)
    if abs(det_M) < 1e-7:
        return False # parallel to triangle
    det_M = 1.0 / det_M
    sad = np.cross(s - a, d)
    u =  np.dot(side2, sad) * det_M
    v = -np.dot(side1, sad) * det_M
    if bias == 'larger':
        nearzero = -1e-6
        nearone = 1.000001
    elif bias == 'smaller':
        nearzero = 1e-6
        nearone = 0.999999
    else:
        raise ValueError('bias must be larger or smaller!')
    if u <= nearzero or v <= nearzero or u + v >= nearone:
        return False # intersection point out of triangle bounds
    s_sign = np.dot(s - a, trinorm)
    e_sign = np.dot(e - a, trinorm)
    if s_sign * e_sign > 0.0:
        return False # intersect point out of segment bounds
    return True

def SegIntersectAnyTriangle(verts, tris, s, e):
    for t in tris:
        if IsectTriSeg([verts[t[0]], verts[t[1]], verts[t[2]]], s, e):
            return True
    return False

def edge_params(verts, tris, edge, ps, pk):
    # Diffracting path over edge. 
    # Coordinate system dz: 0 at v1, increasing towards v2.
    # dzi, dzd, dz0, dz2 are coordinates (displacements) in that system.
    # Coordinate system z: 0 at diffraction apex, increasing towards d side.
    # zi, zd, zedge1, and zedge2 are coordinates (displacements) in that system.
    # The tricky part is that the origins and positive directions may be
    # arbitrarily on either side.
    ei1, ei2 = edge['i1'], edge['i2']
    ev = np.array(verts[ei2]) - np.array(verts[ei1])
    ui, dzi = upright(verts[ei1], verts[ei2], ps, True)
    ud, dzd = upright(verts[ei1], verts[ei2], pk, True)
    ri = norm(ui)
    rd = norm(ud)
    dz = abs(dzd - dzi)
    if ri < 1e-7 or rd < 1e-7:
        raise ValueError('Source/sink almost on edge')
    zi = -dz / ((rd / ri) + 1) #zi < 0
    zd = dz + zi
    li = sqrt(zi * zi + ri * ri)
    ld = sqrt(zd * zd + rd * rd)
    phii = asin(ri / li)
    coordrelsign = -1 if np.sign(dzd - dzi) < 0 else 1
    dz0 = dzi - coordrelsign * zi
    dz2 = norm(ev)
    apex = normalize(ev) * dz0 + np.array(verts[ei1])
    inedge = 0 <= dz0 <= dz2
    zedge1 = -dz0 * coordrelsign
    zedge2 = (dz2 - dz0) * coordrelsign
    zedgelo = min(zedge1, zedge2)
    zedgehi = max(zedge1, zedge2)
    # Angles
    thetaw = edge['angle']
    v = pi / thetaw
    tu1 = triuprightn(verts, tris[edge['tis'][0]], ei1, ei2)
    tu2 = tu1 if len(edge['tis']) == 1 else triuprightn(verts, tris[edge['tis'][1]], ei1, ei2)
    tn1 = np.cross(ev, tu1)
    thetai = acos3(tu1, tn1, normalize(ui))
    thetad = acos3(tu1, tn1, normalize(ud))
    thetaw2 = acos3(tu1, tn1, tu2)
    if thetai <= thetaw2 <= thetad or thetai >= thetaw2 >= thetad:
        raise ValueError('Source and sink on opposite sides of wedge')
    if thetai >= thetaw2 <= thetad:
        # thetaw2 must be the largest, the other angles must be within the wedge
        thetai = tau - thetai
        thetad = tau - thetad
        thetaw2 = tau - thetaw2
    if thetai > thetad:
        # Oops, we picked the wrong triangle as incident
        tu2, tu1 = tu1, tu2
        thetai = thetaw2 - thetai
        thetad = thetaw2 - thetad
    assert(abs(thetaw2 - thetaw) < 1e-3)
    # Done
    return ri, rd, zi, zd, li, ld, phii, apex, inedge, zedgelo, zedgehi, thetai, thetad, v
    
