#Offline diffraction simulation scripts from pisha2020approximate
#Copyright (C) 2020 Louis Pisha <lpisha@eng.ucsd.edu> / Sonic Arts Research & Development <sonicarts.ucsd.edu> / UC San Diego / The Regents of the University of California 
#See LICENSE.txt for licensing information

# Results - Diffraction by a long, thin gap, while BTM is set to ignore
# non-shadowed diffraction. Figure 14 in pisha2020approximate.

import math
import numpy as np
import matplotlib.pyplot as plt
import matplotlib.tri as mtri
from matplotlib.widgets import Slider, CheckButtons, RadioButtons
from mpl_toolkits.mplot3d import Axes3D

from common import *
import examples3d
from xtd import gtd_core, utd_core, call_xtd
from btm import call_btm
from vdat import vdat

# Parameters

sz = 0.03
ps = [-1.0, 0.0, -1.0]
pk = [-1.0, 0.0, 1.0]

c = 345.0
ringdefs = defaultringdefs(9, 64, 44100, c)
vdatparams = defaultvdatparams()

btmloparams = {'fs': 88200, 'c': c, 'irlen': 512, 'totalirlen': 16384, 'goodnumint': False}
btmhiparams = {'fs': 160000, 'c': c, 'irlen': 65536, 'totalirlen': 65536*16, 'goodnumint': True}
btmparams = btmloparams


# Subplots

plt.figure(figsize=(5.0, 5.0), dpi=100)
simax = plt.axes((0.05, 0.55, 0.4, 0.4), projection='3d')
magax = plt.axes((0.55, 0.55, 0.4, 0.4))

# Functions
def points_array_to_xyz(points):
    return [[float(p[i]) for p in points] for i in range(3)]
def plot_mesh(verts, tris):
    if len(tris) == 0: return
    tri = mtri.Triangulation([v[0] for v in verts], [v[1] for v in verts], tris)
    simax.plot_trisurf(tri, [v[2] for v in verts], lw=0.2, edgecolor='black', color='grey', alpha=0.5)
def plot_edges(verts, edges):
    def edge_verts(e):
        return [[verts[e['i1']][d], verts[e['i2']][d]] for d in range(0,3)]
    def vcenter(v):
        return [(vv[0] + vv[1]) / 2.0 for vv in v]
    def vdiff(v):
        return [(vv[1] - vv[0]) for vv in v]
    for e in edges:
        v = edge_verts(e)
        simax.plot(*v, lw=3.0, c='black', marker='')
def plot_xtd_path(apex, inedge):
    simax.plot(*points_array_to_xyz([ps, apex, pk]), lw=1.5, marker='', 
        c='darkgoldenrod' if inedge else 'lightgrey', linestyle='-' if inedge else '--')
    #simax.plot([apex[0]], [apex[1]], [apex[2]], marker='^', c='black' if inedge else 'grey')

# Mesh
verts, tris, edges = examples3d.longgap(sz)

zoom = 0.6
simax.clear()
simax.set_xlim(-zoom, zoom)
simax.set_ylim(-zoom, zoom)
simax.set_zlim(-zoom, zoom)
plot_mesh(verts, tris)
plot_edges(verts, edges)
    
for i, edge in enumerate(edges):
    ri, rd, zi, zd, li, ld, phii, apex, inedge, zedgelo, zedgehi, thetai, thetad, v \
        = edge_params(verts, tris, edge, ps, pk)
    t = tris[edge['tis'][0]]
    if IsectTriSeg([verts[t[0]], verts[t[1]], verts[t[2]]], ps, pk):
        plot_xtd_path(apex, inedge)

btmres_na, btmfreqs, _, _ = call_btm(verts, tris, edges, ps, pk, False, btmparams)
btmres_na = todb(btmres_na)

_, vdatmagres, vdatlens, vdatsps = vdat(
    verts, tris, edges, ps, pk, ringdefs, btmfreqs, True, vdatparams)
vdatmagres = todb(vdatmagres)

simax.plot([ps[0]], [ps[1]], [ps[2]], c='lime', marker='o', markersize=12)
simax.plot([pk[0]], [pk[1]], [pk[2]], c='blue', marker='o', markersize=12)

magax.clear()
magax.set_xscale('log')
magax.set_xlim(20, 25000)
magax.set_ylim(-70, -10)
ring_fs = [r['f'] for r in ringdefs]
magax.plot(btmfreqs, vdatmagres, marker='', linestyle='--', color='darkorange', label='VDaT')
magax.plot(btmfreqs, btmres_na, marker='', color='cyan', label='ED, shadowed only')
magax.set_xlabel('Frequency (Hz)')
#magax.set_ylabel('Amplitude (dB)')
magax.legend(loc='lower left')

plt.show()
