#Offline diffraction simulation scripts from pisha2020approximate
#Copyright (C) 2020 Louis Pisha <lpisha@eng.ucsd.edu> / Sonic Arts Research & Development <sonicarts.ucsd.edu> / UC San Diego / The Regents of the University of California 
#See LICENSE.txt for licensing information

# Results - Diffraction by other objects: an angled plane, a wedge, and a thick
# object. The results for the thick object are computed externally (by
# Svensson's Matlab toolbox) and saved in the text file 'thickresults.txt'.
# Figure 15 in pisha2020approximate.

import math
import numpy as np
import matplotlib.pyplot as plt
import matplotlib.tri as mtri
from matplotlib.widgets import Slider, CheckButtons, RadioButtons
from mpl_toolkits.mplot3d import Axes3D

from common import *
import examples3d
from xtd import gtd_core, utd_core, call_xtd
from btm import call_btm
from vdat import vdat

# Parameters

ps = [-0.4, 0.0, -5.0]
pk = [-0.4, 0.0, 5.0]
sz = 50.0

c = 345.0
ringdefs = defaultringdefs(9, 64, 44100, c)
vdatparams = defaultvdatparams()

btmloparams = {'fs': 88200, 'c': c, 'irlen': 512, 'totalirlen': 16384, 'goodnumint': False}
btmhiparams = {'fs': 160000, 'c': c, 'irlen': 65536, 'totalirlen': 65536*16, 'goodnumint': True}
btmparams = btmhiparams

plt.figure(figsize=(5.0,2.95), dpi=100)
dax = plt.axes((0.05,0.05,0.35,0.9))
ax = plt.axes((0.49,0.18,0.48,0.75))
dax.set_xlim(-6, 2)
dax.set_ylim(-6, 6)
dax.set_axis_off()
ax.set_xscale('log')
ax.set_xlim(20, 20000)
ax.set_ylim(-48, -20)
ax.set_xlabel('Frequency (Hz)')
ax.set_ylabel('Amplitude (dB)')

def get_results(meshname, theta):
    verts, tris, edges = getattr(examples3d, meshname)(theta * pi / 180.0, sz)
    btmres, btmfreqs, _, _ = call_btm(verts, tris, edges, ps, pk, True, btmparams)
    btmres = todb(btmres)
    _, vdatmagres, _, _ = vdat(
        verts, tris, edges, ps, pk, ringdefs, btmfreqs, True, vdatparams)
    vdatmagres = todb(vdatmagres)
    return btmfreqs, btmres, vdatmagres

def plot_results(freqs, baseres, thisres, ls, c, l, mk=''):
    fwlsd = FW_LSD(freqs, baseres, thisres, True)
    if l != 'VDaT':
        l += ': {:.1f} dB'.format(fwlsd)
    ax.plot(freqs, thisres, marker=mk, linestyle=ls, color=c, label=l)

def thickwall():
    thickfreqs = []
    btm_thick = []
    with open('thickresults.txt', 'r') as tr:
        for res in tr:
            toks = [t for t in res.strip().split(' ') if t]
            assert(len(toks) == 2)
            thickfreqs.append(float(toks[0]))
            btm_thick.append(float(toks[1]))
    thickfreqs = np.array(thickfreqs)
    btm_thick = todb(np.array(btm_thick))
    verts, tris, edges = examples3d.halfplane_nonorthog(0.0, sz)
    _, vdatmagres, _, _ = vdat(
        verts, tris, edges, ps, pk, ringdefs, thickfreqs, True, vdatparams)
    vdatmagres = todb(vdatmagres)
    return thickfreqs, btm_thick, vdatmagres

freqs, btm_plane, v = get_results('halfplane_nonorthog', 0.0)
plot_results(freqs, v, v, '-', 'darkorange', 'VDaT', '.')
plot_results(freqs, v, btm_plane, ':', 'blue', 'P')
dax.plot([-sz, 0], [0, 0], linestyle=':', c='blue', marker='')

a = 90.0
ca = sz * cos((a * pi / 180.0) / 2.0)
sa = sz * sin((a * pi / 180.0) / 2.0)
_, btm_angled, _ = get_results('halfplane_nonorthog', a)
plot_results(freqs, v, btm_angled, '-', 'green', 'A')
dax.plot([-ca, 0], [sa, 0], linestyle='-', c='green', marker='', linewidth=3)

_, btm_wedge, _ = get_results('infwedge', a)
plot_results(freqs, v, btm_wedge, '--', 'red', 'W')
dax.plot([-ca, 0, -ca], [sa, 0, -sa], linestyle='--', c='red', marker='')

t = 2.0
thickfreqs, btm_thick, v_thick = thickwall()
plot_results(thickfreqs, v_thick, btm_thick, '-.', 'purple', 'T')
dax.plot([-sz,0,0,-sz],[-t/2,-t/2,t/2,t/2], linestyle='-.', c='purple', marker='')

dax.plot([ps[0], 0, pk[0]], [ps[2], 0, pk[2]], c='darkgoldenrod', marker='')
dax.plot([ps[0]], [ps[2]], c='lime', marker='o', markersize=12)
dax.plot([pk[0]], [pk[2]], c='blue', marker='o', markersize=12)

ax.legend(loc='lower left')
#plt.tight_layout()
plt.show()
